(* Time-stamp: <modified the 12/02/2015 (at 10:45) by Erwan Jahier> *)

open Soc2cIdent
open Data


type var_kind = (* XXX poor names: fixme! *)
  | ML_IO of Soc.key (* For memoryless soc in Heap mode *)
  | M_IO (* *)
  | Local (* for soc local variables *)

let rec (type_to_string : Data.t -> string -> string) = 
  fun v n -> 
    (* in order to print arrays of arrays type size in the good order, 
       we accumulate the array size when v of of type array, and
       we print it otherwise (finish).
    *)
    let rec finish acc str =
      match acc with
        |[] -> str
        | s::ts -> Printf.sprintf "%s[%d]" (finish ts str) s 
    in 
    let rec aux acc v n =
      match v with
        | Bool -> finish acc ("_boolean "^n)
        | Int -> finish acc ("_integer "^n)
        | Real-> finish acc ("_real "^n)
        | Extern s -> finish acc ((id2s s)^" "^n)
        | Enum  (s, sl) -> finish acc (id2s s ^" "^n)
        | Struct (sid,_) -> finish acc ((id2s sid)^" "^n)
        | Array (ty, sz) -> aux (sz::acc) ty n
        | Alpha nb -> finish acc ("alpha_"^(string_of_int nb)^" "^n) 
        | Alias(a,_) -> finish acc (a^" "^n)
    in
    aux [] v n

let string_of_flow_decl (id, t) = 
  Printf.sprintf "   %s;\n" (type_to_string t (id2s id)) 


let rec (data_type_to_c: Data.t  -> string -> string) =
  fun v n -> 
    let rec finish acc str =
      match acc with
        |[] -> str
        | s::ts -> Printf.sprintf "%s[%d]" (finish ts str) s
    in 
    let rec aux acc v n =
      match v with
        | Bool -> finish acc ("_boolean "^n)
        | Int -> finish acc ("_integer "^n)
        | Real-> finish acc ("_real "^n)
        | Extern s -> finish acc ((id2s s)^" "^n)
        | Enum  (s, sl) -> finish acc (id2s s ^" "^n)
        | Struct (sid,_) -> finish acc ((id2s sid)^" "^n)
        | Array (ty, sz) -> aux (sz::acc) ty n
        | Alpha nb -> finish acc ("alpha_"^(string_of_int nb)^" "^n) 
        | Alias(a,_) -> finish acc (a^" "^n)
    in
    aux [] v n

let rec (lic_type_to_c: Lic.type_  -> string -> string) =
  fun t n -> 
    match t with
      | Lic.Struct_type_eff (name, fl) ->
        let field_to_c (id,(tf,_opt)) = 
          Printf.sprintf "\n   %s;" (data_type_to_c (Lic2soc.lic_to_data_type tf) (id2s id)) 
        in
        ((Printf.sprintf "struct { %s\n  }" 
            (String.concat "" (List.map field_to_c fl)))^ " " ^ n) 
      | Lic.Enum_type_eff (name, l) -> "_integer"^ " " ^ n              
      | _ -> type_to_string (Lic2soc.lic_to_data_type t) n


let rec (lic_type_to_c_old: Lic.type_  -> string -> string) =
  fun t n -> match t with
    | Lic.Bool_type_eff -> "_boolean"^ " " ^ n
    | Lic.Int_type_eff  -> "_integer"^ " " ^ n
    | Lic.Real_type_eff -> "_real"^ " " ^ n
    | Lic.External_type_eff (name) -> (long2s name)^ " " ^ n
    | Lic.Abstract_type_eff (name, t) -> lic_type_to_c t n
    | Lic.Enum_type_eff (name, l) -> "_integer"^ " " ^ n
    | Lic.Array_type_eff (ty, sz) -> 
      Printf.sprintf "%s %s[%d]" (lic_type_to_c ty "") n sz
    | Lic.Struct_type_eff (name, fl) ->
      let field_to_c (id,(tf,_opt)) = 
        Printf.sprintf "  %s %s;\n" (lic_type_to_c tf "")  (id2s id)
      in
      (Printf.sprintf "struct %s { %s }" 
        (long2s name) 
        (String.concat "" (List.map field_to_c fl)))^ " " ^ n
    | Lic.TypeVar Lic.Any -> assert false
    | Lic.TypeVar Lic.AnyNum -> assert false


(* exported *)
let (gen_c_switch : string -> (string * string) list -> string) = 
  fun cond_var cases ->
    if Lv6MainArgs.global_opt.Lv6MainArgs.soc2c_no_switch then
      let l = List.map
        (fun (v,code) -> Printf.sprintf "  if (%s == %s) {\n%s\n  } else {\n" cond_var v code) 
        cases
      in
      let switch = String.concat "" l in
      let closing_curly = String.concat "" (List.map (fun _ -> "}") cases) in
      Printf.sprintf "%s  %s" switch closing_curly
    else
        let case_list = List.map
          (fun (v, code) -> Printf.sprintf "  case %s:\n%s  break;\n" v code) cases  
        in
        let cases = String.concat "" case_list in
        Printf.sprintf "  switch (%s){\n%s}\n" cond_var cases
